"""grounding相关的选择器"""
from kele.control.registry import register

from .rule_selector import GroundingRuleSelector
from .term_selector import GroundingFlatTermWithWildCardSelector

__all__ = [
    "GroundingFlatTermWithWildCardSelector",
    "GroundingRuleSelector",
    "register",
]
